from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Dict, List

from qblox_instruments.qcodes_drivers.module import Module
from qblox_instruments.qcodes_drivers.sequencer import Sequencer


@dataclass
class ConnectionConfig:
    bulk_value: List[str] = field(default_factory=list)
    out0: str = None  # Enum: {‘I’, ‘Q’, ‘off’}
    out1: str = None  # Enum: {‘I’, ‘Q’, ‘off’}
    out2: str = None  # Enum: {‘I’, ‘Q’, ‘off’}
    out3: str = None  # Enum: {‘I’, ‘Q’, ‘off’}
    acq: str = None  # Enum: {False, ‘in0’, True, ‘off’}
    acq_I: str = None  # Enum: {‘in0’, ‘in1’, ‘off’}
    acq_Q: str = None  # Enum: {‘in0’, ‘in1’, ‘off’}


@dataclass
class NcoConfig:
    freq: float = None
    phase_offs: float = None
    prop_delay_comp: int = None
    prop_delay_comp_en: bool = None


@dataclass
class AwgConfig:
    cont_mode_en_path0: bool = None
    cont_mode_en_path1: bool = None
    cont_mode_waveform_idx_path0: int = None
    cont_mode_waveform_idx_path1: int = None

    upsample_rate_path0: int = None
    upsample_rate_path1: int = None

    gain_path0: float = None
    gain_path1: float = None

    offset_path0: float = None
    offset_path1: float = None

    mod_en: bool = None


@dataclass
class MixerConfig:
    phase_offset: float = None
    gain_ratio: float = None


@dataclass
class SquareWeightAcq:
    integration_length: int = None


@dataclass
class ThresholdedAcqConfig:
    rotation: int = None
    threshold: float = None
    marker_en: bool = None
    marker_address: int = None
    marker_invert: bool = None
    trigger_en: bool = None
    trigger_address: int = None
    trigger_invert: bool = None


@dataclass
class TtlAcqConfig:
    auto_bin_incr_en: bool = None
    threshold: float = None
    input_select: int = None


@dataclass
class SequencerConfig:
    sync_en: bool = None
    marker_ovr_en: bool = None
    marker_ovr_value: int = None

    trigger_count_thresholds: int = field(default_factory=dict)  # key in [0,15]
    trigger_threshold_inverts: bool = field(default_factory=dict)  # key in [0,15]

    connection: ConnectionConfig = field(default_factory=lambda: ConnectionConfig())
    nco: NcoConfig = field(default_factory=lambda: NcoConfig())
    awg: AwgConfig = field(default_factory=lambda: AwgConfig())
    mixer: MixerConfig = field(default_factory=lambda: MixerConfig())

    demod_en_acq: bool = None
    square_weight_acq: SquareWeightAcq = field(default_factory=lambda: SquareWeightAcq())
    thresholded_acq: ThresholdedAcqConfig = field(
        default_factory=lambda: ThresholdedAcqConfig()
    )
    ttl_acq: TtlAcqConfig = field(default_factory=lambda: TtlAcqConfig())


@dataclass
class OffsetConfig:
    out0: float = None
    out1: float = None
    out2: float = None
    out3: float = None

    in1: float = None
    in0: float = None

    out0_path0: float = None
    out0_path1: float = None
    out1_path0: float = None
    out1_path1: float = None

    in0_path0: float = None
    in0_path1: float = None


@dataclass
class LoConfig:
    out0_en: bool = None
    out0_freq: float = None
    out1_en: bool = None
    out1_freq: float = None

    out0_in0_en: bool = None
    out0_in0_freq: float = None


@dataclass
class AttConfig:
    out0: int = None
    out1: int = None
    in0: int = None


@dataclass
class GainConfig:
    in0: int = None
    in1: int = None


@dataclass
class ScopeAcqConfig:
    sequencer_select: int = None
    trigger_mode_path0: str = None  # Enum: {‘sequencer’, ‘level’}
    trigger_mode_path1: str = None  # Enum: {‘sequencer’, ‘level’}
    trigger_level_path0: float = None
    trigger_level_path1: float = None
    avg_mode_en_path0: bool = None
    avg_mode_en_path1: bool = None


@dataclass
class ModuleConfig:
    marker_inverts: Dict[int, bool] = field(default_factory=dict)

    offset: OffsetConfig = field(default_factory=lambda: OffsetConfig())
    lo: LoConfig = field(default_factory=lambda: LoConfig())
    attenuation: AttConfig = field(default_factory=lambda: AttConfig())
    gain: GainConfig = field(default_factory=lambda: GainConfig())
    scope_acq: ScopeAcqConfig = field(default_factory=lambda: ScopeAcqConfig())


@dataclass
class QbloxConfig:
    slot_idx: int = None
    module: ModuleConfig = field(default_factory=lambda: ModuleConfig())
    sequencers: Dict[int, SequencerConfig] = field(default_factory=dict)


class QbloxConfigHelper(ABC):
    def __init__(self, config: QbloxConfig = None):
        self.config = config or QbloxConfig()

    def configure(self, module: Module, sequencer: Sequencer):
        self.reset(module, sequencer)
        self.configure_module(module, self.config.module)
        self.configure_sequencer(sequencer, self.config.sequencers[sequencer.seq_idx])

    @abstractmethod
    def reset(self, module: Module, sequencer: Sequencer):
        pass

    @abstractmethod
    def configure_module(self, module: Module, config: ModuleConfig):
        pass

    @abstractmethod
    def configure_sequencer(self, sequencer: Sequencer, config: SequencerConfig):
        pass

    def configure_connection(self, sequencer: Sequencer, config: SequencerConfig):
        if config.connection:
            sequencer.connect_sequencer(*config.connection.bulk_value)

    def configure_nco(self, sequencer: Sequencer, config: SequencerConfig):
        if config.nco.freq:
            sequencer.nco_freq(config.nco.freq)
        if config.nco.prop_delay_comp_en:
            sequencer.nco_prop_delay_comp_en(config.nco.prop_delay_comp_en)
        if config.nco.prop_delay_comp:
            sequencer.nco_prop_delay_comp(config.nco.prop_delay_comp)
        if config.nco.phase_offs:
            sequencer.nco_phase_offs(config.nco.phase_offs)

    def configure_awg(self, sequencer: Sequencer, config: SequencerConfig):
        if config.awg.mod_en:
            sequencer.mod_en_awg(config.awg.mod_en)
        if config.awg.gain_path0:
            sequencer.gain_awg_path0(config.awg.gain_path0)
        if config.awg.gain_path1:
            sequencer.gain_awg_path1(config.awg.gain_path1)
        if config.awg.offset_path0:
            sequencer.offset_awg_path0(config.awg.offset_path0)
        if config.awg.offset_path1:
            sequencer.offset_awg_path1(config.awg.offset_path1)

        if config.awg.cont_mode_en_path0:
            sequencer.cont_mode_en_awg_path0(config.awg.cont_mode_en_path0)
        if config.awg.cont_mode_en_path1:
            sequencer.cont_mode_en_awg_path1(config.awg.cont_mode_en_path1)
        if config.awg.cont_mode_waveform_idx_path0:
            sequencer.cont_mode_waveform_idx_awg_path0(
                config.awg.cont_mode_waveform_idx_path0
            )
        if config.awg.cont_mode_waveform_idx_path1:
            sequencer.cont_mode_waveform_idx_awg_path1(
                config.awg.cont_mode_waveform_idx_path1
            )

        if config.awg.cont_mode_en_path0:
            sequencer.cont_mode_en_awg_path0(config.awg.cont_mode_en_path0)
        if config.awg.cont_mode_en_path1:
            sequencer.cont_mode_en_awg_path1(config.awg.cont_mode_en_path1)
        if config.awg.cont_mode_waveform_idx_path0:
            sequencer.cont_mode_waveform_idx_awg_path0(
                config.awg.cont_mode_waveform_idx_path0
            )
        if config.awg.cont_mode_waveform_idx_path1:
            sequencer.cont_mode_waveform_idx_awg_path1(
                config.awg.cont_mode_waveform_idx_path1
            )

    def configure_mixer(self, sequencer: Sequencer, config: SequencerConfig):
        if config.mixer.phase_offset:
            sequencer.mixer_corr_phase_offset_degree(config.mixer.phase_offset)
        if config.mixer.gain_ratio:
            sequencer.mixer_corr_gain_ratio(config.mixer.gain_ratio)

    def calibrate_mixer(self, module: Module, sequencer: Sequencer):
        """
        - Compensates for the LO leakage for input(s)/output(s) depending on module
        - Suppresses undesired sideband on sequencer
        """
        self.calibrate_lo_leakage(module)
        self.calibrate_sideband(sequencer)

    @abstractmethod
    def calibrate_lo_leakage(self, module):
        pass

    def calibrate_sideband(self, sequencer):
        sequencer.sideband_cal()
        sequencer.arm_sequencer()
        sequencer.start_sequencer()


class QcmConfigHelper(QbloxConfigHelper):
    def reset(self, module: Module, sequencer: Sequencer):
        module.disconnect_outputs()


class QcmRfConfigHelper(QcmConfigHelper):
    def configure_module(self, module, config):
        self.configure_lo(module, config)
        self.configure_attenuation(module, config)
        self.configure_offset(module, config)

    def configure_sequencer(self, sequencer, config):
        if config.sync_en:
            sequencer.sync_en(config.sync_en)
        self.configure_connection(sequencer, config)
        self.configure_nco(sequencer, config)
        self.configure_awg(sequencer, config)
        self.configure_mixer(sequencer, config)

    def configure_lo(self, module, config):
        if config.lo.out0_en:
            module.out0_lo_en(config.lo.out0_en)  # Switch the LO 0 on
        if config.lo.out0_freq:
            module.out0_lo_freq(config.lo.out0_freq)
        if config.lo.out1_en:
            module.out1_lo_en(config.lo.out1_en)  # Switch the LO 1 on
        if config.lo.out1_freq:
            module.out1_lo_freq(config.lo.out1_freq)

    def configure_attenuation(self, module, config):
        if config.attenuation.out0:
            module.out0_att(config.attenuation.out0)
        if config.attenuation.out1:
            module.out1_att(config.attenuation.out1)

    def configure_offset(self, module, config):
        if config.offset.out0_path0:
            module.out0_offset_path0(config.offset.out0_path0)
        if config.offset.out0_path1:
            module.out0_offset_path1(config.offset.out0_path1)
        if config.offset.out1_path0:
            module.out1_offset_path0(config.offset.out1_path0)
        if config.offset.out1_path1:
            module.out1_offset_path1(config.offset.out1_path1)

    def calibrate_lo_leakage(self, module):
        module.out0_lo_cal()
        module.out1_lo_cal()


class QrmConfigHelper(QbloxConfigHelper):
    def configure(self, module, sequencer):
        super().configure(module, sequencer)

        # TODO - make these settings dynamic
        module.scope_acq_sequencer_select(sequencer.seq_idx)
        module.scope_acq_trigger_mode_path0("sequencer")
        module.scope_acq_trigger_mode_path1("sequencer")

    def reset(self, module: Module, sequencer: Sequencer):
        module.disconnect_outputs()
        module.disconnect_inputs()

    def configure_scope_acq(self, module, config):
        scope_acq = config.scope_acq
        if scope_acq.avg_mode_en_path0:
            module.scope_acq_avg_mode_en_path0(scope_acq.avg_mode_en_path0)
        if scope_acq.avg_mode_en_path1:
            module.scope_acq_avg_mode_en_path1(scope_acq.avg_mode_en_path1)


class QrmRfConfigHelper(QrmConfigHelper):
    def configure_module(self, module, config):
        self.configure_lo(module, config)
        self.configure_attenuation(module, config)
        self.configure_offset(module, config)
        self.configure_scope_acq(module, config)

    def configure_sequencer(self, sequencer, config):
        if config.sync_en:
            sequencer.sync_en(config.sync_en)
        self.configure_connection(sequencer, config)
        self.configure_nco(sequencer, config)
        self.configure_awg(sequencer, config)
        self.configure_mixer(sequencer, config)
        self.configure_acq(sequencer, config)

    def configure_lo(self, module, config):
        if config.lo.out0_in0_en:
            module.out0_in0_lo_en(config.lo.out0_in0_en)  # Switch the LO on
        if config.lo.out0_in0_freq:
            module.out0_in0_lo_freq(config.lo.out0_in0_freq)

    def configure_attenuation(self, module, config):
        if config.attenuation.out0:
            module.out0_att(config.attenuation.out0)
        if config.attenuation.in0:
            module.in0_att(config.attenuation.in0)

    def configure_offset(self, module, config):
        if config.offset.out0_path0:
            module.out0_offset_path0(config.offset.out0_path0)
        if config.offset.out0_path1:
            module.out0_offset_path1(config.offset.out0_path1)
        if config.offset.in0_path0:
            module.in0_offset_path0(config.offset.in0_path0)
        if config.offset.in0_path1:
            module.in0_offset_path1(config.offset.in0_path1)

    def configure_acq(self, sequencer, config):
        # Square weight integration
        if config.demod_en_acq:
            sequencer.demod_en_acq(config.demod_en_acq)
        if config.square_weight_acq.integration_length:
            sequencer.integration_length_acq(config.square_weight_acq.integration_length)

    def calibrate_lo_leakage(self, module):
        module.out0_in0_lo_cal()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
import os
import shutil
from pathlib import Path

from qat.purr.utils.logger import get_default_logger


class QatCache:
    """
    Folder control object for the various caches of the toolchain. Consolidates
    creation/deletion and validation.
    """

    def __init__(self, root_folder=None):
        if root_folder is None:
            root_folder = os.path.realpath(os.path.join(__file__, "..", "..", "..", ".."))

        self.qat_root = os.path.join(root_folder, ".qat")
        self.ll_cache = os.path.realpath(os.path.join(self.qat_root, "ll"))
        self.qs_cache = os.path.realpath(os.path.join(self.qat_root, "qs"))
        self.qat_cache = os.path.realpath(os.path.join(self.qat_root, "qat"))

    def create_cache_folders(self):
        """Creates folders if they don't exist."""
        Path(self.ll_cache).mkdir(parents=True, exist_ok=True)
        Path(self.qs_cache).mkdir(parents=True, exist_ok=True)
        Path(self.qat_cache).mkdir(parents=True, exist_ok=True)

    def delete_cache_folders(self):
        try:
            shutil.rmtree(self.qat_root)
        except OSError as e:
            get_default_logger().warn(
                f"Attempted to delete cache folders at {self.qat_root}, "
                f"threw exception. {str(e)}"
            )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd

import qiskit_experiments.library.randomized_benchmarking.standard_rb as rb
from qiskit import qasm2

from qat.purr.compiler.runtime import get_builder
from qat.purr.integrations.qasm import Qasm2Parser
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


def randomized_benchmarking(hardware, nseeds, *args, **kwargs):
    """
    Generate randomized benchmarks. Please see Qiskits ``randomized_benchmarking_seq``
    for argument documentation. Returns a dictionary holding the QuantumProcess that
    relates to the various benchmarking runs and a list of the sequence lengths.
    """
    # Due to the variable return value we can't directly unpack.
    lengths = kwargs.pop("lengths", [1, 2, 4])
    physical_qubits = kwargs.pop("physical_qubits", [0])

    results = dict()
    index = 0
    for seed in range(nseeds):
        circuit_list = []
        qiskit_results = rb.StandardRB(
            physical_qubits=physical_qubits,
            lengths=lengths,
            num_samples=1,
            seed=seed,
            *args,
            **kwargs,
        )
        circuits = qiskit_results.circuits()
        for circuit in circuits:
            qasm = qasm2.dumps(circuit)
            circuit_list.append(Qasm2Parser().parse(get_builder(hardware), qasm))
        results[index] = circuit_list
        index = index + 1

    return results, lengths



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
import atexit
import logging
import os
import shutil
import sys
from datetime import date, datetime
from enum import Enum
from logging.config import dictConfig
from pathlib import Path
from typing import IO, List, Union

from numpy import savetxt

from qat.purr.utils.serializer import json_dump, json_load

# Formatted to "[INFO] 2020-08-25 19:54:28,216 (module_name.function_name:line_number) - message"
default_logger_format = "[%(levelname)s] %(asctime)s - %(name)s - (%(module)s.%(funcName)s:%(lineno)d) - %(message)s"
json_format = (
    '{"level": "%(levelname)s", "time": "%(asctime)s",'
    '"module name": "%(module)s", "function name": "%(funcName)s", "line number": %(lineno)d,'
    '"message": "%(message)s"},'
)


class LoggerLevel(Enum):
    CRITICAL = logging.CRITICAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    OUTPUT = logging.NOTSET + 2
    CODE = logging.NOTSET + 1
    NOTSET = logging.NOTSET

    def __repr__(self):
        return self.name


class BasicLogger(logging.Logger):
    """
    The basic logger class that should be used. Upon setup, this is provided to the
    built-in logging by calling ``logging.setLoggerClass``. This way, every new logger
    created will be of this class. This allows us to define custom fields and functions
    that we want to use with our loggers. This class should not be instantiated
    separately, only call ``logging.getLogger("qat.purr.some_name")``, and this will return
    an instance of :class:`BasicLogger`.
    """

    def __init__(self, name: str, _log_folder: "LogFolder" = None):
        logging.Logger.__init__(self, name)
        self.setLevel(logging.INFO)
        self.log_folder = _log_folder or LogFolder()

    @property
    def logs_path(self):
        return self.log_folder.folder_path

    def output(
        self,
        data,
        *args,
        cell_type: str = None,
        fit_type: str = None,
        msg: str = "",
        section_level=1,
    ):
        """
        Displays the result of some experiment or computation. This logging function is
        PuRR specific.

        The output depends on the :paramref:`cell_type` parameter:

        - ``table`` - it creates a table in markdown syntax. The data needs to be a list
          of lists of strings, where the first list will be the header of the table, and
          then each list represents a row.
        - ``fit`` - it displays the Fit function in LaTeX syntax. It also requires the
          :paramref:`fit_type` parameter
        - ``new_section`` - creates a new section in markdown syntax. The title will be
          the string provided by :paramref:`data`

        :param data: The data to output based on the rest of the parameters.
        :param cell_type: It can have multiple values: ``table``, ``fit``,
            ``new_section``
        :param fit_type: Required when :paramref:`cell_type` is ``fit``. It can be one
            of the following types: ``SINUSOID``, ``SINUSOID_EXPONENTIAL``,
            ``POLYNOMIAL``, ``EXPONENTIAL``, ``LORENTZIAN``
        :param msg: Message to insert in front of the table if :paramref:`cell_type` is
            ``table``
        :param section_level: The level of the new section in markdown, if
            :paramref:`cell_type` is ``new_section``
        """
        logging.Logger.log(
            self,
            level=LoggerLevel.OUTPUT.value,
            msg=data,
            *args,
            extra={
                "cell_type": cell_type,
                "fit_type": fit_type,
                "note": msg,
                "section_level": section_level,
            },
        )

    def code(self, source: List[str]):
        """
        Outputs a section of executable code (used for Jupyter files). This logging
        function is PuRR specific.

        :param source: The list of code lines as strings
        """
        logging.Logger.log(self, level=LoggerLevel.CODE.value, msg="\n".join(source))

    def save_object(self, obj, name: str, numpy_arr: bool = False):
        """
        Serializes the specified object. This logging function is PuRR specific.

        :param obj: The object to be serialized.
        :param name: A name must be also provided, which will be used as the file name
                     and in case of Jupyter loggers, also as the name of the variable
                     which will contain the loaded object.
        """
        if numpy_arr:
            code = [
                "# load the measured data from the file",
                "from numpy import loadtxt",
                f"{name} = loadtxt(r'{name}.txt')",
            ]
            logging.Logger.log(self, level=LoggerLevel.CODE.value, msg="\n".join(code))
            return
        file_name = f"{name}.json"
        file_path = os.path.join(self.logs_path, file_name)
        with open(file_path, "w") as f:
            json_dump(obj, f, indent=4, ensure_ascii=False)

        # After serializing the object above, a code snippet will be added to the
        # Jupyter notebook to load the object from the JSON, and save it to a variable.
        # This will allow the user to actively analyse the saved object after running
        # the Jupyter script.
        code = [
            "from qat.purr.logger import *",
            f"{name} = load_data('{os.path.relpath(file_path, self.logs_path)}')",
        ]
        logging.Logger.log(self, level=LoggerLevel.CODE.value, msg="\n".join(code))

    def close(self):
        """Closes this logger, cleans up the file handles and appropriate folder."""
        for handler in self.handlers:
            try:
                handler.close()
            except Exception as e:
                print(f"Logger handler failed to close cleanly. Message: {e}")

        if self.log_folder is not None:
            try:
                self.log_folder.close()
            except Exception as e:
                print(f"Log folder failed to close cleanly. Message: {e}")

    _dummy_log = logging.makeLogRecord({})
    record_override_key = "$_enable_record_override"

    def makeRecord(
        self,
        name,
        level,
        fn,
        lno,
        msg,
        args,
        exc_info,
        func=None,
        extra: dict = None,
        sinfo=None,
    ):
        """
        Override that allows us to override record values via the extras dictionary.
        Initially built to allow for printing out messages that look like they come from
        different places in the source code than where the logger was called.
        """
        if extra is None:
            extra = {}

        if extra.pop(self.record_override_key, False):
            # Strip off values that would conflict with the makeRecord validation then
            # just apply them afterwards.
            overwriting_values = {
                key: value
                for key, value in extra.items()
                if key in self._dummy_log.__dict__
            }
            for key in overwriting_values:
                del extra[key]

            # If we have overrides for the function values, just apply them.
            if "fn" in extra:
                fn = extra.pop("fn")

            if "lno" in extra:
                lno = extra.pop("lno")

            if "func" in extra:
                func = extra.pop("func")
        else:
            overwriting_values = {}

        record = super().makeRecord(
            name, level, fn, lno, msg, args, exc_info, func, extra, sinfo
        )
        for key, value in overwriting_values.items():
            record.__dict__[key] = value

        return record

    def __enter__(self):
        pass

    def __exit__(self, exc, value, tb):
        self.close()

    def __del__(self):
        self.close()


class ConsoleLoggerHandler(logging.StreamHandler):
    """
    Basic console handler for the logger. It defaults to stdout.
    """

    def __init__(self, stream: IO = sys.stdout):
        super().__init__(stream)
        self.setFormatter(logging.Formatter(default_logger_format))

    def __repr__(self):
        return "Console logger handler"


class FileLoggerHandler(logging.FileHandler):
    """
    Basic file handler for the logger. A file path must be provided. The log file is
    created with a delay, so the stream is None until the first emit. This also allows
    to write some initial stuff to the log file when creating it.
    """

    def __init__(self, file_path: str):
        super().__init__(os.path.abspath(file_path), mode="w", delay=True)
        self.setFormatter(logging.Formatter(default_logger_format))

    def emit(self, record):
        if self.stream is None:
            os.makedirs(os.path.dirname(self.baseFilename), exist_ok=True)
            self.stream = self._open()
            self.create_initial_file()
        logging.FileHandler.emit(self, record)

    def create_initial_file(self):
        """
        Implement this method in the derived class to insert some initial text in the
        log file. Use emit and flush while writing directly to the stream.
        """
        pass

    def __repr__(self):
        return "File logger handler (path: %s)" % self.baseFilename


class JsonHandler(FileLoggerHandler):
    """
    The JSON file handler needed for logging in JSON format. In
    :class:`logging.FileHandler`, at each time something is written to the file, the
    emit function is called. By overriding the method, the JSON format can be ensured at
    each writing.
    """

    def __init__(self, file_path):
        super().__init__(file_path)
        self.file_terminator = "{}]}" + self.terminator

    def create_initial_file(self):
        pass

    def emit(self, record):
        # TODO: Unicode problem
        if self.stream is None:
            os.makedirs(os.path.dirname(self.baseFilename), exist_ok=True)
            self.stream = self._open()
            self.create_initial_file()
        self.stream.seek(0, os.SEEK_END)
        self.stream.seek(self.stream.tell() - len(self.file_terminator) - 1, os.SEEK_SET)
        FileLoggerHandler.emit(self, record)
        self.stream.write(self.file_terminator)
        self.flush()


class JsonLoggerHandler(JsonHandler):
    """
    The basic JSON file handler logger. It is intended to generate the same output as
    :class:`FileLoggerHandler`, but in JSON format.
    """

    def __init__(self, file_path):
        extension = Path(file_path).suffix
        if not extension == ".json":
            raise ValueError("Only JSON file paths are accepted!")
        super().__init__(file_path)
        self.setFormatter(logging.Formatter(json_format))

    def create_initial_file(self):
        self.stream.write(
            "{" + f'{os.linesep}"entries": [{os.linesep}' + "{}]}" + self.terminator
        )
        self.flush()

    def __repr__(self):
        return "JSON console logger handler (path: %s)" % self.baseFilename


class CompositeLogger(BasicLogger):
    """
    The default logger class of PuRR. It is intended to store all the loggers in a list,
    and when logging, the functions from here should be called, which iterate through
    the list and apply the logging to each logger separately. This way, only one
    function needs to be called when logging, and it is ensured, that all the enabled
    loggers will log the message.
    """

    def __init__(
        self,
        loggers_or_names: List[Union[str, logging.Logger]] = None,
        _log_folder=None,
    ):
        """Creates the list of loggers on which the logging functions will iterate

        :param logger_names: List of loggers by their names
            (e.g. ``["qat.purr.json", "qat.purr.file"]``) or actual logger instances.
        """
        super().__init__("default", _log_folder)

        self.loggers = []
        if loggers_or_names is None:
            loggers_or_names = []

        # Add root to our composite so consumers can also hook into our logs.
        root = logging.getLogger()
        loggers_or_names.append(root)
        self.add_loggers(loggers_or_names)

        # Set the root to the lowest non-custom log level activated.
        root.setLevel(
            min([val.level for val in self.loggers if (float(val.level / 10)).is_integer()])
        )

    def add_loggers(self, loggers_or_names: List[Union[str, logging.Logger]] = ()):
        if loggers_or_names is not None:
            for val in loggers_or_names:
                if isinstance(val, str):
                    self.loggers.append(logging.getLogger(val))
                elif isinstance(val, logging.Logger):
                    self.loggers.append(val)

    def _add_stack_levels(self, kwargs):
        """
        Due to the way the loggers work, we need to go back up the stack a few calls to
        get the real caller.
        """
        kwargs["stacklevel"] = kwargs.get("stacklevel", 0) + 2
        return kwargs

    def info(self, msg: str, *args, **kwargs):
        kwargs = self._add_stack_levels(kwargs)
        for logger in self.loggers:
            logger.info(msg, *args, **kwargs)

    def debug(self, msg: str, *args, **kwargs):
        kwargs = self._add_stack_levels(kwargs)
        for logger in self.loggers:
            logger.debug(msg, *args, **kwargs)

    def error(self, msg: str, *args, **kwargs):
        kwargs = self._add_stack_levels(kwargs)
        for logger in self.loggers:
            logger.error(msg, *args, **kwargs)

    def warning(self, msg: str, *args, **kwargs):
        kwargs = self._add_stack_levels(kwargs)
        for logger in self.loggers:
            logger.warning(msg, *args, **kwargs)

    def critical(self, msg: str, *args, **kwargs):
        kwargs = self._add_stack_levels(kwargs)
        for logger in self.loggers:
            logger.critical(msg, *args, **kwargs)

    def log(self, level, msg: str, *args, **kwargs):
        kwargs = self._add_stack_levels(kwargs)
        if isinstance(level, LoggerLevel):
            level = level.value
        for logger in self.loggers:
            logger.log(level, msg, *args, **kwargs)

    def exception(self, msg: str, *args, **kwargs):
        kwargs = self._add_stack_levels(kwargs)
        for logger in self.loggers:
            logger.exception(msg, *args, **kwargs)

    def output(self, data, **kwargs):
        for logger in self.loggers:
            if isinstance(logger, BasicLogger):
                logger.output(data, **kwargs)

    def code(self, source: List[str]):
        for logger in self.loggers:
            if isinstance(logger, BasicLogger):
                logger.code(source)

    def save_object(self, obj, name: str, numpy_arr: bool = False, overwrite=False):
        """
        Iterates through the list of loggers and calls the corresponding save_object
        implementations for each of them. This is the method that should be called from
        the code.
        """
        if numpy_arr:
            if not overwrite:
                i = 2
                original_name = name
                while os.path.exists(
                    os.path.join(self.log_folder.folder_path, f"{name}.txt")
                ):
                    name = f"{original_name}_{i}"
                    i += 1
            file_path = os.path.join(self.log_folder.folder_path, f"{name}.txt")
            savetxt(file_path, obj)
        for logger in self.loggers:
            if isinstance(logger, BasicLogger):
                logger.save_object(obj, name, numpy_arr=numpy_arr)
        return name

    def close(self):
        super(CompositeLogger, self).close()
        for logger in self.loggers:
            if isinstance(logger, BasicLogger):
                logger.close()


class LogFolder:
    """
    It is the main log folder in which all the log files are saved. It can be configured
    multiple ways, like the base folder path, which can be a specified path of the disk
    or ``None`` to save the logs in the system temporary folder. The
    :paramref:`labber_style` specifies whether to create a Labber-style folder hierarchy
    for the logs or not. If not, depending on the :paramref:`folder_name` parameter, it
    will either create random folders for each run, or if :paramref:`folder_name` is not
    ``None``, it will create a sub-folder specified by :paramref:`folder_name`. Also, a
    :paramref:`prefix` and :paramref:`suffix` can be specified to append the created log
    folder (if :paramref:`labber_style` is not True).
    """

    def __init__(
        self,
        base_folder_path: str = None,
        labber_style: bool = None,
        cleanup: bool = None,
        folder_name: str = None,
        prefix: str = None,
        suffix: str = None,
    ):
        """
        The constructor for the LogFolder. It can be configured by the parameters. If
        the parameters are not provided, the default variables are used from the front
        of the file (which can be also set by importing a configuration).

        :param base_folder_path: Specifies the base directory, where the new log folder
            will be created. If it is ``None``, then it is set to the default value
            ``default_logger_base_directory``, which is set by the imported
            configuration file, otherwise it is defined at the top of this module. If
            the default value is ``None``, the log folder will be created in the
            system's TMP folder.
        :param labber_style: If it is true, it will create a labber hierarchy log
            folder.
        :param cleanup: If it is true, it will remove the log folder together with the
            logs at the end of execution.
        :param folder_name: If :paramref:`labber_style` is false, then
            :paramref:`folder_name` will be the name of the new log folder instead of
            generating a random one.
        :param prefix: It appends to the front of the generated folder name
        :param suffix: It appends to the end of the generated folder name
        """
        self.starting_date = date.today()
        self.needs_cleanup: bool = cleanup

        if base_folder_path is None:
            base_folder_path = "logs"

        self.base_folder_path = os.path.abspath(base_folder_path)
        os.makedirs(self.base_folder_path, exist_ok=True)

        folder_name = folder_name or ""
        if prefix is not None:
            folder_name = f"{prefix}_{folder_name}"
        if suffix is not None:
            folder_name = f"{folder_name}_{suffix}"
        if labber_style:
            self.folder_path = self.create_sub_folder_labber_style(
                base_folder_path, folder_name=folder_name
            )
        else:
            self.folder_path = os.path.join(base_folder_path, folder_name)
            os.makedirs(self.folder_path, exist_ok=True)

        self.folder_path = os.path.abspath(self.folder_path)

    def get_log_file_path(self, file_name: str = "log", over_write=True):
        file_path = os.path.join(self.folder_path, file_name)
        if not over_write:
            if os.path.exists(file_path):
                raise OSError("File already exists!")
        return file_path

    def create_sub_folder_labber_style(self, base_folder, folder_name: str = None):
        if folder_name is None or folder_name == "":
            folder_name = ""
        else:
            folder_name = folder_name + "."
        now = datetime.now()
        folder_name = folder_name + f"{now.hour:02d}.{now.minute:02d}.{now.second:02d}"
        main_folder_path = self.get_main_folder_path_labber_style(base_folder)
        original_name = folder_name
        i = 2
        while os.path.exists(os.path.join(main_folder_path, folder_name)):
            folder_name = f"{original_name}_{i}"
            i += 1
        final_sub_folder_path = os.path.join(main_folder_path, folder_name)
        os.makedirs(final_sub_folder_path, exist_ok=True)
        return final_sub_folder_path

    @staticmethod
    def get_main_folder_path_labber_style(base_folder):
        now = datetime.now()
        year, month, day = now.year, now.month, now.day
        main_folder_path = os.path.join(
            base_folder, f"{year:04d}", f"{month:02d}", f"Data_{month:02d}{day:02d}"
        )
        return main_folder_path

    def close(self):
        if self.folder_path is not None and os.path.isdir(self.folder_path):
            if self.needs_cleanup:
                shutil.rmtree(self.base_folder_path)
            else:
                try:
                    os.removedirs(self.base_folder_path)
                except OSError:
                    pass

    def __repr__(self):
        return self.folder_path

    def __del__(self):
        self.close()


class KeywordFilter(logging.Filter):
    """
    A customized keyword filter that can be added to a log handler or a logger. Filters
    all the log messages, and if the message content contains the keyword, the log will
    not be printed.
    """

    def __init__(self, keyword=""):
        super().__init__(keyword)
        self.keyword = keyword

    def filter(self, record: logging.LogRecord):
        if self.keyword in record.msg:
            return False
        return True


class ModuleFilter(logging.Filter):
    """
    A customized module filter that can be added to a log handler or a logger. Filters
    all the log messages, and if the log was produced by a module with the specified
    module name, the log will not pass.
    """

    def __init__(self, module_name=""):
        super().__init__(module_name)
        self.module_name = module_name

    def filter(self, record: logging.LogRecord):
        if self.module_name in record.module:
            return False
        return True


class LevelFilter(logging.Filter):
    """
    Filter out the debug messages from the Jupyter logs. This is needed because the
    specialized logging functions, like code or output have smaller level than the
    DEBUG logging level (so that other than Jupyter handlers don't process them).
    """

    def __init__(self, level):
        super().__init__()
        self.level = logging.getLevelName(level)

    def filter(self, record: logging.LogRecord):
        if isinstance(self.level, int) and record.levelno == self.level:
            return False
        return True


logging.setLoggerClass(BasicLogger)
"""
These specialized logging levels are registered with the logging system, so they can be
retrieved by using for example logging.getLevelName('CODE').
"""
logging.addLevelName(LoggerLevel.OUTPUT.value, "OUTPUT")
logging.addLevelName(LoggerLevel.CODE.value, "CODE")


def import_logger_configuration(logger_config: dict, log_folder: LogFolder = None):
    """
    It imports the configuration of the loggers from a JSON data structure. This must be
    in the format described by `logging.config
    <https://docs.python.org/3/library/logging.config.html>`_ built-in module.

    It can also contain some additional settings:

    - :default_logger_directory: This is where a new log folder will be created for each
        execution. If it is set to None the system's temp folder is used.
    - :default_logger_cleanup: Specifies whether the log folders should be removed after
        execution or not.
    - :default_logger_labber_style: If this is true, it will create a log folder
        hierarchy in labber style at the specified **default_logger_directory**

    The logger list in the config file may also contain some additional settings:

    - :class: If this is specified, then the logger is of a custom class, not included
        in the built-in logging package. Similar to how the handlers are defined by the
        '()' key if they are custom handlers.
    - :active: If this is false, than the corresponding logger will not be imported.
        It's an easier way not to include a logger than to remove it from the config
        file, because then if the logger will be needed sometime, it doesn't require to
        re-write the config file, just change **active** from false to true.

    The configuration may also contain the starting log folder settings
    (:paramref:`log_folder`). Each time the logging configuration is imported, the log
    folder will be set up as it is specified. If this is not provided in the JSON
    structure, than the created log folder will use the default settings (which can also
    be specified in the configuration, as described above).


    :param logger_config: The JSON data structure from the logger_settings.json
                          configuration
    :param log_folder: The log folder to be used instead of the one specified in the
                       configuration file
    :return: List of the imported loggers. They are already loaded and configured.
    """
    if log_folder is None:
        if "log_folder" in logger_config:
            log_folder = LogFolder(**logger_config["log_folder"])
        else:
            log_folder = LogFolder()

    if "external_loggers" in logger_config:
        logger_config["loggers"].update(logger_config["external_loggers"])

    non_active_loggers = []
    for key, value in logger_config["loggers"].items():
        if "active" in value and value["active"] == 0:
            non_active_loggers.append(key)

    for key, value in logger_config["handlers"].items():
        # If we have a path append the current log folder onto it.
        file_path = value.get("file_path", None)
        if file_path is not None:
            value["file_path"] = os.path.join(log_folder.folder_path, file_path)

    for key in non_active_loggers:
        logger_config["loggers"].pop(key)

    dictConfig(logger_config)
    log_keys = set(logger_config["loggers"].keys())
    if "external_loggers" in logger_config:
        log_keys.difference_update(set(logger_config["external_loggers"].keys()))

    return CompositeLogger(list(log_keys), log_folder)


def get_logger_config(config_file=None, log_folder: LogFolder = None):
    """
    It imports the logger configuration from the provided JSON file. If this is not
    provided, then the current directory is searched for a logger_settings.json
    configuration file. If not found, then the default JSON file is read from
    qat/purr/logger_settings.json

    :param config_file: The path to the JSON file on the disk containing the logger
                        configuration
    :param log_folder: The log folder to be used instead of the one specified in the
                       configuration file
    :return: A DefaultLogger instance configured with the names of the imported loggers
    """

    if config_file is None:
        config_file = "logger_settings.json"

    potential_file = config_file
    if not os.path.isfile(potential_file):
        potential_file = os.path.join(os.getcwd(), config_file)
        if not os.path.isfile(potential_file):
            potential_file = os.path.join(os.path.dirname(__file__), config_file)

    if not os.path.isfile(potential_file):
        print(
            f"Log config file {config_file} doesn't exist and can't be found using "
            "default search patterns. Loading default configuration."
        )
        potential_file = os.path.join(os.path.dirname(__file__), "logger_settings.json")

    with open(potential_file, "r") as f:
        logger_config = json_load(f)
        return import_logger_configuration(logger_config, log_folder)


@atexit.register
def close_logger():
    """
    This method is executed upon exit, and it closes all the file handlers from the
    default loggers.

    This allows to remove the log folder after the execution is finished. This is needed
    because otherwise, the handlers would be closed after everything else, after the log
    folder is being removed.
    """
    def_logger = get_default_logger()
    if def_logger is not None:
        def_logger.close()


def load_object_from_log_folder(file_path: str):
    """
    Loads and deserializes an object from its JSON representation from the disk.

    :param file_path: The JSON file on the disk
    :return: The loaded object after deserialization
    """
    extension = Path(file_path).suffix
    if not extension == ".json":
        raise ValueError("Only JSON file paths are accepted!")
    with open(
        os.path.join(get_default_logger().log_folder.folder_path, file_path), "r"
    ) as f:
        obj = json_load(f)
    get_default_logger().debug(f"Object loaded: {str(obj)}")
    return obj


def save_object_to_log_folder(obj, sub_folder_path: str):
    """
    Serializes the specified object.
    """
    extension = Path(sub_folder_path).suffix
    if not extension == ".json":
        raise ValueError("Only JSON file paths are accepted!")
    sub_folder_path = os.path.join(
        get_default_logger().log_folder.folder_path, sub_folder_path
    )
    dir_name = os.path.dirname(sub_folder_path)
    if not os.path.exists(dir_name):
        os.makedirs(dir_name)

    with open(sub_folder_path, "w") as f:
        json_dump(obj, f, indent=4, ensure_ascii=False)

    return sub_folder_path


_default_logging_instance = None


def get_default_logger():
    """
    Initializes the global logger or fetches one if it already exists.
    """
    global _default_logging_instance
    if _default_logging_instance is None:
        _default_logging_instance = get_logger_config()

        # Do a very hacky check about whether we're in a testing environment and force
        # clean-up if we are.
        try:
            import traceback

            stack = traceback.extract_stack()
            is_test_env = any(
                val.filename is not None and val.filename.endswith(f"unittest\\loader.py")
                for val in stack
            )
            if is_test_env:
                _default_logging_instance.log_folder.needs_cleanup = True
        except:
            _default_logging_instance.warning("Test environment detection failed.")

    return _default_logging_instance



from contextlib import ExitStack, contextmanager
from dataclasses import dataclass, field
from typing import Dict, List

import numpy as np

from qat.purr.backends.qblox.config import SequencerConfig
from qat.purr.backends.qblox.constants import Constants
from qat.purr.backends.qblox.ir import Opcode, Sequence, SequenceBuilder
from qat.purr.backends.utilities import evaluate_shape
from qat.purr.compiler.devices import PulseChannel, PulseShapeType
from qat.purr.compiler.emitter import QatFile
from qat.purr.compiler.instructions import (
    Acquire,
    CustomPulse,
    Delay,
    FrequencyShift,
    Id,
    MeasurePulse,
    PhaseReset,
    PhaseShift,
    PostProcessing,
    Pulse,
    QuantumInstruction,
    Repeat,
    Synchronize,
    Waveform,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


def get_nco_phase_arguments(phase_rad: float) -> int:
    phase_deg = np.rad2deg(phase_rad)
    phase_deg %= 360
    return round(phase_deg * Constants.NCO_PHASE_STEPS_PER_DEG)


def get_nco_set_frequency_arguments(frequency_hz: float) -> int:
    frequency_steps = round(frequency_hz * Constants.NCO_FREQ_STEPS_PER_HZ)

    if (
        frequency_steps < -Constants.NCO_FREQ_LIMIT_STEPS
        or frequency_steps > Constants.NCO_FREQ_LIMIT_STEPS
    ):
        min_max_frequency_in_hz = (
            Constants.NCO_FREQ_LIMIT_STEPS / Constants.NCO_FREQ_STEPS_PER_HZ
        )
        raise ValueError(
            f"NCO frequency must be in [-{min_max_frequency_in_hz:e}, {min_max_frequency_in_hz:e}] Hz. "
            f"Got {frequency_hz:e} Hz"
        )

    return frequency_steps


def calculate_duration(instruction, return_samples: bool = True):
    """
    Calculates the duration of the instruction. The duration is either in nanoseconds
    or in number of samples.
    """
    target_channels = [
        target for target in instruction.quantum_targets if isinstance(target, PulseChannel)
    ]
    if not any(target_channels):
        return 0

    # TODO: Allow for multiple pulse channel targets.
    if len(target_channels) > 1 and not isinstance(instruction, PhaseReset):
        log.warning(
            f"Attempted to calculate duration of {str(instruction)} that has multiple"
            f" target channels. We're arbitrarily using the duration of the first channel "
            f"to calculate instruction duration."
        )

    pc = target_channels[0].physical_channel
    block_size = pc.block_size
    block_time = pc.block_time
    block_number = np.ceil(
        round(instruction.duration / block_time, 4)
    )  # round to remove floating point errors
    if return_samples:
        calc_sample = block_number * block_size
    else:
        calc_sample = block_number * block_time

    return calc_sample


@dataclass
class QbloxPackage:
    target: PulseChannel = None
    sequence: Sequence = None
    sequencer_config: SequencerConfig = field(default_factory=lambda: SequencerConfig())
    timeline: np.ndarray = None


class QbloxEmitter:
    def emit(self, qat_file: QatFile) -> List[QbloxPackage]:
        contexts: Dict[PulseChannel, QbloxContext] = {}

        with ExitStack() as stack:
            inst_iter = iter(qat_file.instructions)
            while (inst := next(inst_iter, None)) is not None:
                if not isinstance(inst, QuantumInstruction):
                    continue  # Ignore classical instructions

                if isinstance(inst, PostProcessing):
                    continue  # Ignore postprocessing

                if isinstance(inst, Synchronize):
                    for target in inst.quantum_targets:
                        if not isinstance(target, PulseChannel):
                            raise ValueError(f"{target} is not a PulseChannel")
                        contexts.setdefault(
                            target, stack.enter_context(QbloxContext(qat_file.repeat))
                        )
                    QbloxContext.synchronize(inst, contexts)
                    continue

                if isinstance(inst, PhaseReset):
                    QbloxContext.reset_phase(inst, contexts)
                    continue

                for target in inst.quantum_targets:
                    if not isinstance(target, PulseChannel):
                        raise ValueError(f"{target} is not a PulseChannel")

                    context = contexts.setdefault(
                        target, stack.enter_context(QbloxContext(qat_file.repeat))
                    )

                    if isinstance(inst, MeasurePulse):
                        acquire = next(inst_iter, None)
                        if acquire is None or not isinstance(acquire, Acquire):
                            raise ValueError(
                                "Found a MeasurePulse but no Acquire instruction followed"
                            )
                        context.measure_acquire(inst, acquire, target)
                    elif isinstance(inst, Waveform):
                        context.waveform(inst, target)
                    elif isinstance(inst, Delay):
                        context.delay(inst)
                    elif isinstance(inst, PhaseShift):
                        context.shift_phase(inst)
                    elif isinstance(inst, FrequencyShift):
                        context.shift_frequency(inst, target)
                    elif isinstance(inst, Id):
                        context.id()

        contexts = self.optimize(contexts)

        return [context.create_package(target) for target, context in contexts.items()]

    def optimize(self, qblox_contexts: Dict) -> Dict:
        # Remove empty contexts
        qblox_contexts = {
            target: context
            for target, context in qblox_contexts.items()
            if not context.is_empty()
        }

        # Remove Opcode.WAIT_SYNC instructions when the experiment contains only a singleton context
        if len(qblox_contexts) == 1:
            context = list(qblox_contexts.values())[0]
            context.sequence_builder.q1asm_instructions = [
                inst
                for inst in context.sequence_builder.q1asm_instructions
                if not inst.opcode == Opcode.WAIT_SYNC
            ]

        return qblox_contexts


class QbloxContext:
    def __init__(self, repeat: Repeat):
        self.sequence_builder = SequenceBuilder()
        self.sequencer_config = SequencerConfig()

        self._repeat_count = repeat.repeat_count
        self._repeat_period = repeat.repetition_period
        self._repeat_reg = None
        self._repeat_label = None

        self._duration: int = 0
        self._timeline: np.ndarray = np.empty(0)

        self._num_hw_avg = 1  # Technically disabled
        self._wf_memory: int = Constants.MAX_SAMPLE_SIZE_WAVEFORMS
        self._wf_index: int = 0
        self._acq_index: int = 0

        self._frequency: float = 0.0  # Keeps record of the frequency shift on the target

        self._registers: List[str] = sorted(
            f"R{index}" for index in range(Constants.NUMBER_OF_REGISTERS)
        )
        self._label_counters = {}

    def _reg_alloc(self) -> str:
        if len(self._registers) < 1:
            raise IndexError(
                "Out of registers. Attempting to use more registers "
                "than available in the Q1 sequence processor"
            )
        return self._registers.pop(0)

    def _reg_free(self, register: str) -> None:
        if register in self._registers:
            raise RuntimeError(f"Cannot free register '{register}' as it's not in use")
        self._registers.append(register)

    def _generate_label(self, prefix: str):
        counter = self._label_counters.setdefault(prefix, 0)
        self._label_counters[prefix] = counter + 1
        return f"{prefix}_{counter}"

    @contextmanager
    def _loop(self, label: str, iter_count: int = 1):
        """
        This context can be used to avoid unrolling loops but needs manual addition of the 0 iteration
        either immediately after the yield or immediately after the context has exited.
        """
        register = self._reg_alloc()
        label = self._generate_label(label)
        self.sequence_builder.move(iter_count, register)
        self.sequence_builder.nop()
        self.sequence_builder.label(label)
        yield register
        self.sequence_builder.loop(register, label)
        self._reg_free(register)

    @contextmanager
    def _inc_jlt(self, label: str, iter_count: int = 1, start=0, step=1):
        """
        The `loop` instruction first subtracts 1 from the register and then exits if that is equal to 0.
        This means the effective range will be [iter_count-1, 1], 0 is not included. This context uses
        the `jlt` jump instruction as it's more flexible than `loop`.

        It allows to incrementally iterate `iter_count` times from `start` (inclusive).
        """
        register = self._reg_alloc()
        label = self._generate_label(label)
        self.sequence_builder.move(start, register)
        self.sequence_builder.nop()
        self.sequence_builder.label(label)
        yield register
        self.sequence_builder.add(register, step, register)
        self.sequence_builder.nop()
        self.sequence_builder.jlt(register, iter_count, label)
        self._reg_free(register)

    def __enter__(self):
        self.clear()
        self._repeat_reg = self._reg_alloc()
        self._repeat_label = self._generate_label("repeat_label")

        self.sequence_builder.set_mrk(3)
        self.sequence_builder.upd_param(Constants.GRID_TIME)
        self.sequence_builder.move(0, self._repeat_reg)
        self.sequence_builder.nop()
        self.sequence_builder.label(self._repeat_label)
        self.sequence_builder.reset_ph()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._wait_seconds(self._repeat_period)
        self.sequence_builder.add(self._repeat_reg, 1, self._repeat_reg)
        self.sequence_builder.nop()
        self.sequence_builder.jlt(self._repeat_reg, self._repeat_count, self._repeat_label)
        self.sequence_builder.stop()

        self._reg_free(self._repeat_reg)
        self._repeat_label = None
        self._repeat_reg = None

    @property
    def duration(self):
        return self._duration

    def is_empty(self):
        """
        Masks away yet-to-be-supported second-state, cancellation, and cross cancellation targets
        This is temporary and criteria will change with more features coming in
        """
        return not (
            self.sequence_builder.waveforms
            or self.sequence_builder.acquisitions
            or [
                inst
                for inst in self.sequence_builder.q1asm_instructions
                if inst.opcode
                in [
                    Opcode.PLAY,
                    Opcode.SET_AWG_OFFSET,
                    Opcode.SET_AWG_GAIN,
                    Opcode.ACQUIRE,
                    Opcode.ACQUIRE_TTL,
                    Opcode.ACQUIRE_WEIGHED,
                ]
            ]
        )

    def clear(self):
        self.sequence_builder.waveforms.clear()
        self.sequence_builder.acquisitions.clear()
        self.sequence_builder.weights.clear()
        self.sequence_builder.q1asm_instructions.clear()

    def create_package(self, target: PulseChannel):
        sequence = self.sequence_builder.build()
        return QbloxPackage(target, sequence, self.sequencer_config, self._timeline)

    def _wait_seconds(self, duration: float):
        self._wait_nanoseconds(int(duration * 1e9))

    def _wait_nanoseconds(self, duration: int):
        if duration <= 0:
            return

        quotient = duration // Constants.MAX_WAIT_TIME
        remainder = duration % Constants.MAX_WAIT_TIME
        if quotient > Constants.LOOP_UNROLL_THRESHOLD:
            with self._loop("wait_label", quotient):
                self.sequence_builder.wait(Constants.MAX_WAIT_TIME)
        elif quotient >= 1:
            for _ in range(quotient):
                self.sequence_builder.wait(Constants.MAX_WAIT_TIME)

        if remainder > 0:
            self.sequence_builder.wait(remainder)

    def _evaluate_waveform(self, waveform: Waveform, target: PulseChannel):
        """
        The waveform is evaluated as a 1d complex array. In QBlox, the Real and Imag parts of the pulse
        represent the digital offset on the AWG. They both must be within range [-1, 1].
        """

        num_samples = int(calculate_duration(waveform, return_samples=True))
        if num_samples < Constants.GRID_TIME:
            if num_samples == 0:
                log.warning(f"Empty pulse.")
            else:
                log.warning(
                    f"""
                    Minimum pulse width is {Constants.GRID_TIME} ns with a resolution of {1} ns.
                    Please round up the width to at least {Constants.GRID_TIME} nanoseconds.
                    """
                )
            return None

        dt = target.sample_time
        length = num_samples * dt
        centre = length / 2.0
        t = np.linspace(
            start=-centre + 0.5 * dt, stop=length - centre - 0.5 * dt, num=num_samples
        )
        pulse = evaluate_shape(waveform, t)
        scale = target.scale
        if isinstance(waveform, (Pulse, CustomPulse)) and waveform.ignore_channel_scale:
            scale = 1

        pulse *= scale
        pulse += target.bias

        if ((pulse.real < -1) | (pulse.real > 1)).any():
            raise ValueError(
                "Voltage range for I exceeded. Make sure all I values are within the range [-1, 1]"
            )
        if ((pulse.imag < -1) | (pulse.imag > 1)).any():
            raise ValueError(
                "Voltage range for Q exceeded. Make sure all I values are within the range [-1, 1]"
            )

        if ((pulse.real > 0.3) | (pulse.imag > 0.3)).any():
            log.warning(
                """
                Values above 0.3 will overdrive the mixer and  produce intermodulation distortion.
                Consider adjusting attenuation instead.
                """
            )

        return pulse

    def _register_waveform(self, measure, target, data):
        num_samples = data.size
        if 2 * num_samples > self._wf_memory:  # for both I and Q
            raise ValueError(
                f"No more waveform memory left for pulse {measure} on channel {target}"
            )

        name = type(measure).__name__
        if isinstance(measure, Pulse):
            name = measure.shape.name

        index = len(self.sequence_builder.waveforms)
        i_index, q_index = self._wf_index, self._wf_index + 1
        self.sequence_builder.add_waveform(f"{name}_{index}_I", i_index, data.real.tolist())
        self.sequence_builder.add_waveform(f"{name}_{index}_Q", q_index, data.imag.tolist())

        self._wf_memory = self._wf_memory - data.size
        self._wf_index = self._wf_index + 2

        return i_index, q_index

    def _register_acquisition(self, acquire: Acquire):
        acq_index = self._acq_index
        self.sequence_builder.add_acquisition(
            acquire.output_variable, acq_index, self._repeat_count
        )
        self._acq_index = acq_index + 1
        return acq_index

    def _binned_acquisition(
        self, delay, acq_index, num_samples, pulse_shape, i_steps, q_steps, i_index, q_index
    ):
        capped_num_samples = min(num_samples, Constants.MAX_WAIT_TIME)
        flight_nanos = int(delay * 1e9)
        capped_flight_nanos = min(flight_nanos, Constants.MAX_WAIT_TIME)
        if pulse_shape == PulseShapeType.SQUARE:
            self.sequence_builder.set_awg_offs(i_steps, q_steps)
            self._wait_seconds(delay)
            self.sequence_builder.acquire(
                acq_index,
                self._repeat_reg,
                capped_num_samples,
            )
            self._wait_seconds((num_samples - capped_num_samples) / 1e9)
            self.sequence_builder.set_awg_offs(0, 0)
            self.sequence_builder.upd_param(Constants.GRID_TIME)
        else:
            self.sequence_builder.play(i_index, q_index, capped_flight_nanos)
            self._wait_seconds((flight_nanos - capped_flight_nanos) / 1e9)
            self.sequence_builder.acquire(
                acq_index,
                self._repeat_reg,
                capped_num_samples,
            )
            self._wait_seconds((num_samples - capped_num_samples) / 1e9)

    def id(self):
        self.sequence_builder.nop()

    def delay(self, inst: Delay):
        self._wait_seconds(inst.duration)
        self._duration = self._duration + inst.duration
        num_samples = int(calculate_duration(inst))
        self._timeline = np.append(self._timeline, [0] * num_samples)

    def waveform(self, waveform: Waveform, target: PulseChannel):
        pulse = self._evaluate_waveform(waveform, target)
        if pulse is None:
            log.warning("This pulse will be ignored.")
            return

        num_samples = pulse.size
        max_duration = min(num_samples, Constants.MAX_WAIT_TIME)
        if isinstance(waveform, Pulse) and waveform.shape == PulseShapeType.SQUARE:
            i_offs_steps = int(
                pulse[0].real * (Constants.MAX_OFFSET_SIZE // 2)  # Signed integer
            )
            q_offs_steps = int(
                pulse[0].imag * (Constants.MAX_OFFSET_SIZE // 2)  # Signed integer
            )
            self.sequence_builder.set_awg_offs(i_offs_steps, q_offs_steps)
            self.sequence_builder.upd_param(max_duration)
            self._wait_seconds((num_samples - max_duration) / 1e9)
            self.sequence_builder.set_awg_offs(0, 0)
            self.sequence_builder.upd_param(Constants.GRID_TIME)
        else:
            i_index, q_index = self._register_waveform(waveform, target, pulse)
            self.sequence_builder.play(i_index, q_index, max_duration)
            self._wait_seconds((num_samples - max_duration) / 1e9)

        self._duration = self._duration + waveform.duration
        self._timeline = np.append(self._timeline, pulse)

    def measure_acquire(
        self, measure: MeasurePulse, acquire: Acquire, target: PulseChannel
    ):
        pulse = self._evaluate_waveform(measure, target)
        if pulse is None:
            log.warning("This pulse will be ignored.")
            return

        acq_index = self._register_acquisition(acquire)
        self.sequencer_config.square_weight_acq.integration_length = calculate_duration(
            acquire
        )

        i_steps, q_steps = None, None
        i_index, q_index = None, None
        if measure.shape == PulseShapeType.SQUARE:
            i_steps = int(pulse[0].real * (Constants.MAX_OFFSET_SIZE // 2))
            q_steps = int(pulse[0].imag * (Constants.MAX_OFFSET_SIZE // 2))
        else:
            i_index, q_index = self._register_waveform(measure, target, pulse)

        if self._num_hw_avg <= Constants.LOOP_UNROLL_THRESHOLD:
            for _ in range(self._num_hw_avg):
                self._binned_acquisition(
                    acquire.delay,
                    acq_index,
                    pulse.size,
                    measure.shape,
                    i_steps,
                    q_steps,
                    i_index,
                    q_index,
                )
        else:
            with self._loop("avg_label", self._num_hw_avg + 1):
                self._binned_acquisition(
                    acquire.delay,
                    acq_index,
                    pulse.size,
                    measure.shape,
                    i_steps,
                    q_steps,
                    i_index,
                    q_index,
                )

        self._duration = self._duration + measure.duration
        self._timeline = np.append(self._timeline, pulse)

    @staticmethod
    def synchronize(inst: Synchronize, contexts: Dict):
        max_duration = max([cxt.duration for cxt in contexts.values()])
        for target in inst.quantum_targets:
            cxt = contexts[target]
            delay_time = max_duration - cxt.duration
            cxt.delay(Delay(target, delay_time))
            cxt.sequence_builder.wait_sync(Constants.GRID_TIME)

    @staticmethod
    def reset_phase(inst: PhaseReset, contexts: Dict):
        for target in inst.quantum_targets:
            cxt = contexts[target]
            cxt._phase = 0.0
            cxt.sequence_builder.reset_ph()
            cxt.sequence_builder.upd_param(Constants.GRID_TIME)

    def shift_phase(self, inst: PhaseShift):
        value = get_nco_phase_arguments(inst.phase)
        self.sequence_builder.set_ph_delta(value)
        self.sequence_builder.upd_param(Constants.GRID_TIME)

    def shift_frequency(self, inst, target):
        # TODO - discuss w/t quantum_target.fixed_if is True or False
        old_frequency = target.frequency + self._frequency
        new_frequency = old_frequency + inst.frequency
        if new_frequency < target.min_frequency or new_frequency > target.max_frequency:
            raise ValueError(
                f"Cannot shift pulse channel frequency from '{old_frequency}' to '{new_frequency}'"
            )

        self._frequency = self._frequency + inst.frequency
        shifted_nco_freq = target.baseband_if_frequency + self._frequency
        value = get_nco_set_frequency_arguments(shifted_nco_freq)  # 1 MHZ <-> 4e6
        self.sequence_builder.set_freq(value)
        self.sequence_builder.upd_param(Constants.GRID_TIME)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
from __future__ import annotations

import re
from copy import deepcopy
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, Set, Union

import numpy as np

from qat.purr.compiler.config import InlineResultsProcessing
from qat.purr.compiler.devices import PulseChannel, PulseShapeType, QuantumComponent, Qubit

if TYPE_CHECKING:
    pass


def _stringify_qubits(qubits):
    return ",".join([str(qb) for qb in qubits])


class PostProcessType(Enum):
    DOWN_CONVERT = "down_convert"
    MEAN = "mean"
    LINEAR_MAP_COMPLEX_TO_REAL = "linear_map_real"
    DISCRIMINATE = "discriminate"
    MUL = "mul"

    def __repr__(self):
        return self.name


class ProcessAxis(Enum):
    TIME = "time"
    SEQUENCE = "sequence"

    def __repr__(self):
        return self.name


class Instruction: ...


class QuantumMetadata(Instruction):
    pass


class QuantumInstruction(Instruction):
    """
    Any node that deals particularly with quantum operations. All quantum operations
    must have some sort of target on the quantum computer, such as a qubit, channel, or
    another form of component.
    """

    def __init__(
        self, quantum_targets: Union[QuantumComponent, List[QuantumComponent]] = None
    ):
        if quantum_targets is None:
            quantum_targets = []
        elif not isinstance(quantum_targets, List):
            quantum_targets = [quantum_targets]

        invalid_targets = [
            val for val in quantum_targets if not isinstance(val, QuantumComponent)
        ]
        if any(invalid_targets):
            invalid_targets_str = ",".join([str(val) for val in invalid_targets])
            raise ValueError(f"Invalid targets for component: {invalid_targets_str}")

        # Quick way to make sure the targets are unique.
        self.quantum_targets: List[QuantumComponent] = []
        for tar in quantum_targets:
            if tar not in self.quantum_targets:
                self.quantum_targets.append(tar)

    @property
    def duration(self):
        return 0.0


def _add_channels(
    instruction: QuantumInstruction,
    channels: Union[Qubit, PulseChannel, List[Union[Qubit, PulseChannel]]],
):
    """Returns the union of `channels` and `instruction`'s quantum targets"""
    if channels is None:
        return instruction.quantum_targets

    if not isinstance(channels, List):
        channels = [channels]

    unique_targets = set(instruction.quantum_targets)
    for target in (
        chan
        for val in channels
        for chan in (val.get_all_channels() if isinstance(val, Qubit) else [val])
    ):
        if not isinstance(target, PulseChannel):
            raise ValueError(
                f"Attempted to try and add non PulseChannel ({target}) to a sync."
            )
        unique_targets.add(target)

    return sorted(list(unique_targets), key=lambda x: x.full_id())


class Repeat(Instruction):
    """
    Global meta-instruction that applies to the entire list of instructions. Repeat
    value of the current operations, also known as shots.
    """

    def __init__(self, repeat_count, repetition_period=None):
        super().__init__()
        self.repeat_count = repeat_count
        self.repetition_period = repetition_period

    def __repr__(self):
        return f"repeat {self.repeat_count},{self.repetition_period}"


class PhaseShift(QuantumInstruction):
    def __init__(self, channel: "PulseChannel", phase: float):
        super().__init__(channel)
        self.phase: float = phase

    @property
    def channel(self) -> "PulseChannel":
        return self.quantum_targets[0]

    def __repr__(self):
        return f"phaseshift {self.channel},{self.phase}"


class FrequencyShift(QuantumInstruction):
    def __init__(self, channel: "PulseChannel", frequency: float):
        super().__init__(channel)
        self.frequency: float = frequency

    @property
    def channel(self) -> "PulseChannel":
        return self.quantum_targets[0]

    def __repr__(self):
        return f"frequencyshift {self.channel},{self.frequency}"


class Id(QuantumInstruction):
    """Simply a no-op, called an Identity gate."""

    def __repr__(self):
        return "id"


class AcquireMode(Enum):
    RAW = "raw"
    SCOPE = "scope"
    INTEGRATOR = "integrator"

    def __repr__(self):
        return self.name


class Delay(QuantumInstruction):
    def __init__(self, quantum_target, time: float):
        super().__init__(quantum_target)
        self.time: float = time

    @property
    def duration(self):
        return self.time

    def __repr__(self):
        return f"delay {str(self.time)}"


class Synchronize(QuantumInstruction):
    """
    Tells the QPU to wait for all the related channels to be free before continuing
    execution on any of them.
    """

    quantum_targets: List[PulseChannel]

    def __init__(
        self,
        sync_channels: Union[Qubit, PulseChannel, List[Union[Qubit, PulseChannel]]],
    ):
        super().__init__()
        self.add_channels(sync_channels)

    def add_channels(
        self, sync_channels: Union[Qubit, PulseChannel, List[Union[Qubit, PulseChannel]]]
    ):
        self.quantum_targets = _add_channels(self, sync_channels)

    def __add__(self, other):
        other_targets = other.quantum_targets if isinstance(other, Synchronize) else other
        new_sync = Synchronize(self.quantum_targets)
        new_sync.add_channels(other_targets)
        return new_sync

    def __repr__(self):
        return f"sync {','.join(name.id for name in self.quantum_targets)}"


class Assign(Instruction):
    """
    Assigns the variable 'x' the value 'y'. This can be performed as a part of running
    on the QPU or by a post-processing pass.
    """

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __repr__(self):
        return f"{self.name} = {str(self.value)}"


class Waveform(QuantumInstruction):
    @property
    def channel(self) -> "PulseChannel":
        return self.quantum_targets[0]


class CustomPulse(Waveform):
    """
    Send a pulse down this particular channel.
    """

    def __init__(
        self,
        quantum_target: "PulseChannel",
        samples: List[np.complex],
        ignore_channel_scale: bool = False,
    ):
        super().__init__(quantum_target)
        self.samples: List[np.complex] = samples
        self.ignore_channel_scale: bool = ignore_channel_scale

    @property
    def duration(self):
        return len(self.samples) * self.channel.sample_time

    def __repr__(self):
        id_ = self.channel.full_id()
        duration = self.duration
        return f"custom pulse {id_},{duration}"


class Pulse(Waveform):
    """
    Send a pulse down this particular channel.
    """

    def __init__(
        self,
        quantum_target: "PulseChannel",
        shape: PulseShapeType,
        width: float,
        amp: float = 1.0,
        phase: float = 0.0,
        drag: float = 0.0,
        rise=0.0,
        amp_setup: float = 0.0,
        scale_factor: float = 1.0,
        zero_at_edges: int = 0,
        beta: float = 0.0,
        frequency: float = 0.0,
        internal_phase: float = 0.0,
        std_dev: float = 0.0,
        square_width: float = 0.0,
        ignore_channel_scale: bool = False,
    ):
        super().__init__(quantum_target)
        self.shape = shape
        self.width = width
        self.amp = amp
        self.phase = phase
        self.drag = drag
        self.rise = rise
        self.amp_setup = amp_setup
        self.scale_factor = scale_factor
        self.zero_at_edges = bool(zero_at_edges)
        self.beta = beta
        self.frequency = frequency
        self.internal_phase = internal_phase
        self.std_dev = std_dev
        self.square_width = square_width
        self.ignore_channel_scale = ignore_channel_scale

    @property
    def duration(self):
        return self.width

    def __repr__(self):
        return (
            f"pulse {self.channel.full_id()},{self.shape.value},{self.amp},"
            f"{self.phase},{self.width},{self.drag},{self.rise}"
        )


class MeasurePulse(Pulse):
    pass


class DrivePulse(Pulse):
    pass


class SecondStatePulse(Pulse):
    pass


class CrossResonancePulse(Pulse):
    pass


class CrossResonanceCancelPulse(Pulse):
    pass


class Acquire(QuantumComponent, QuantumInstruction):
    suffix_incrementor: int = 0

    def __init__(
        self,
        channel: "PulseChannel",
        time: float = None,
        mode: AcquireMode = None,
        output_variable=None,
        existing_names: Set[str] = None,
        delay=None,
        filter: Union[Pulse, CustomPulse] = None,
    ):
        super().__init__(channel.full_id())
        super(QuantumComponent, self).__init__(channel)
        self.time: float = time or 1.0e-6
        self.mode: AcquireMode = mode or AcquireMode.RAW
        self.delay = delay
        self.output_variable = output_variable or self.generate_name(existing_names)
        self.filter: Union[Pulse, CustomPulse] = self._check_filter(filter)

    def _check_filter(self, filter):
        if filter:
            for target in self.quantum_targets:
                if isinstance(target, PulseChannel):
                    dt = target.physical_channel.sample_time
                    if not np.isclose(filter.duration, dt * (self.time // dt), atol=1e-12):
                        raise ValueError(
                            f"Filter duration '{filter.duration}' must be equal to Acquire "
                            f"duration '{self.time}'."
                        )
        return filter

    def generate_name(self, existing_names=None):
        return build_generated_name(existing_names, f"{self.channel.id}")

    @property
    def duration(self):
        return self.time

    @property
    def channel(self) -> "PulseChannel":
        return next(iter(self.quantum_targets), None)

    def __repr__(self):
        out_var = f"->{self.output_variable}" if self.output_variable is not None else ""
        mode = f",{self.mode.value}" if self.mode is not None else ""
        return f"acquire {self.channel.full_id()},{self.time}{mode}{out_var}"


class PostProcessing(QuantumInstruction):
    """
    States what post-processing should happen after data has been acquired. This can
    happen in the FPGA's or a software post-process.
    """

    def __init__(self, acquire: Acquire, process, axes=None, args=None):
        super().__init__(acquire)
        if axes is not None and not isinstance(axes, List):
            axes = [axes]

        self.process: PostProcessType = process

        # TODO: Should find out why ags can't reference model objects in builder serialization.
        #   But deep copying them, as the values shouldn't vary, should be fine.
        self.args: List = deepcopy(args) or []
        self.axes: List[ProcessAxis] = axes or []
        self.output_variable = acquire.output_variable
        self.result_needed = False

    @property
    def acquire(self) -> Acquire:
        return self.quantum_targets[0]

    def __repr__(self):
        axis = ",".join([axi.value for axi in self.axes])
        args = f",{','.join(str(arg) for arg in self.args)}" if len(self.args) > 0 else ","
        output_var = f"->{self.output_variable}" if self.output_variable is not None else ""
        return (
            f"{self.process.value} {self.acquire.output_variable}{args}{axis}{output_var}"
        )


class Reset(QuantumInstruction):
    """Resets this qubit to its starting state."""

    def __init__(self, qubit: Union[List[Qubit], Qubit]):
        if not isinstance(qubit, List):
            qubit = [qubit]

        invalid_reset_targets = [
            str(val) for val in qubit if not isinstance(val, (Qubit, PulseChannel))
        ]
        if any(invalid_reset_targets):
            raise ValueError(
                "Tried to reset on non-qubit/pulse channel "
                f"{', '.join(invalid_reset_targets)}."
            )

        super().__init__(
            [val.get_drive_channel() if isinstance(val, Qubit) else val for val in qubit]
        )

    def __repr__(self):
        return f"reset {','.join([str(qb) for qb in self.quantum_targets])}"


class PhaseReset(QuantumInstruction):
    """
    Reset the phase shift of all the channels
    """

    quantum_targets: List[PulseChannel]

    def __init__(
        self,
        reset_channels: Union[Qubit, PulseChannel, List[Union[Qubit, PulseChannel]]],
    ):
        super().__init__()
        self.add_channels(reset_channels)

    def add_channels(
        self,
        reset_channels: Union[Qubit, PulseChannel, List[Union[Qubit, PulseChannel]]],
    ):
        self.quantum_targets = _add_channels(self, reset_channels)

    def __add__(self, other):
        other_targets = other.quantum_targets if isinstance(other, PhaseReset) else other
        new_reset = PhaseReset(self.quantum_targets)
        new_reset.add_channels(other_targets)
        return new_reset

    def __repr__(self):
        return f"phase reset {','.join(name.id for name in self.quantum_targets)}"


class Return(Instruction):
    """A statement defining what to return from a quantum execution."""

    def __init__(self, variables: List[str] = None):
        if variables is None:
            variables = []

        if not isinstance(variables, List):
            variables = [variables]

        self.variables = variables

    def __repr__(self):
        return f"return {','.join(self.variables)}"


class SweepOperation:
    """Common parent for all things that need differentiating during a sweep."""

    pass


class SweepValue(SweepOperation):
    def __init__(self, name, value):
        self.name = name
        self.value = value


class DeviceUpdate(QuantumInstruction):
    """
    Dynamically assigns a value to a particular symbol or hardware attribute during
    execution.

    .. note:: It's still unknown how this will be represented in the instructions themselves, but that'll come later.
    For now we perform programatic modification and a before/after state.
    """

    def __init__(self, target: QuantumComponent, attribute: str, value):
        super().__init__()
        self.target = target
        self.attribute = attribute
        self.value = value

    def __repr__(self):
        return f"{self.target.full_id()}.{self.attribute} = {str(self.value)}"


class Sweep(Instruction):
    """
    This is a global meta-instruction that performs a sweep over values, effectively
    performing a loop over the instructions replacing a variable with a specific value
    per time.

    Nested sweeps are run in the order they're added and are performed after repeats. So
    a 1000 repeat with a 4 sweep followed by a 2 will run a total of 8000 iterations.
    """

    def __init__(self, operations: Union[SweepValue, List[SweepValue]] = None):
        super().__init__()

        if operations is None:
            operations = []
        elif not isinstance(operations, List):
            operations = [operations]

        self.variables: Dict[str, List[Any]] = {op.name: op.value for op in operations}

        # Get the length of the variables, which we will then assume is the sweep
        # length.
        sweep_lengths = [len(value) for value in self.variables.values()]
        if len(set(sweep_lengths)) > 1:
            raise ValueError("Sweep variables have inconsistent lengths.")

    @property
    def length(self):
        return next(iter([len(value) for value in self.variables.values()]), 0)

    def __repr__(self):
        args = ",".join(key + "=" + str(value) for key, value in self.variables.items())
        return f"sweep {args}"


class Jump(Instruction):
    """
    Classic jump instruction, should be linked to label with an optional condition.
    """

    def __init__(self, label: Union[str, Label], condition=None):
        self.condition = condition
        if isinstance(label, Label):
            self.target = label.name
        else:
            self.target = label

    def __repr__(self):
        if self.condition is not None:
            return f"if {str(self.condition)} -> {str(self.target)}"
        else:
            return f"-> {str(self.target)}"


class BinaryOperator:
    """Binary operator, such as ``x == y``, ``x != y`` etc."""

    def __init__(self, left, right):
        self.left = left
        self.right = right


class Equals(BinaryOperator):
    def __repr__(self):
        return f"{str(self.left)} == {str(self.right)}"


class NotEquals(BinaryOperator):
    def __repr__(self):
        return f"{str(self.left)} != {str(self.right)}"


class GreaterThan(BinaryOperator):
    def __repr__(self):
        return f"{str(self.left)} > {str(self.right)}"


class GreaterOrEqualThan(BinaryOperator):
    def __repr__(self):
        return f"{str(self.left)} >= {str(self.right)}"


class LessThan(BinaryOperator):
    def __repr__(self):
        return f"{str(self.left)} < {str(self.right)}"


class LessOrEqualThan(BinaryOperator):
    def __repr__(self):
        return f"{str(self.left)} <= {str(self.right)}"


def is_generated_name(name: str):
    return re.match(".*generated_name_[0-9]*$", name) is not None


def build_generated_name(existing_names=None, prefix=None):
    if existing_names is None:
        existing_names = set()

    if prefix is None:
        prefix = ""

    if any(prefix) and not prefix.endswith("_"):
        prefix = f"{prefix}_"

    variable_name = f"{prefix}generated_name_{np.random.randint(np.iinfo(np.int32).max)}"
    while variable_name in existing_names:
        variable_name = (
            f"{prefix}generated_name_{np.random.randint(np.iinfo(np.int32).max)}"
        )

    existing_names.add(variable_name)
    return variable_name


class Variable:
    """
    States that this value is actually a variable that should be fetched instead.
    """

    def __init__(self, name, var_type=None, value=None):
        self.name = name
        self.var_type = var_type
        self.value = value

    @staticmethod
    def with_random_name(existing_names=None, var_type=None, value=None):
        return Variable(Variable.generate_name(existing_names), var_type, value)

    @staticmethod
    def generate_name(existing_names=None):
        return build_generated_name(existing_names)

    def __repr__(self):
        return self.name


class Label(Instruction):
    """
    Label to apply to a line of code. Used as anchors for other instructions like jumps.
    """

    def __init__(self, name):
        """If you need a name, use generate_name and pass in existing values."""
        self.name = name

    @staticmethod
    def with_random_name(existing_names=None):
        return Label(Label.generate_name(existing_names))

    @staticmethod
    def generate_name(existing_names=None):
        return build_generated_name(existing_names)

    def __repr__(self):
        return f"{self.name}:"


class IndexAccessor(Variable):
    """Used to access an array index on a particular variable."""

    def __init__(self, name, index):
        super().__init__(name)
        self.index = index

    def __repr__(self):
        return f"{self.name}[{self.index}]"


class ResultsProcessing(Instruction):
    def __init__(self, variable: str, res_processing: InlineResultsProcessing):
        self.variable = variable
        self.results_processing = res_processing

    def __repr__(self):
        return f"{self.variable}: {str(self.results_processing.name)}"



import abc
from typing import Dict

import numpy as np

from qat.purr.compiler.config import ErrorMitigationConfig
from qat.purr.compiler.hardware_models import QuantumHardwareModel


class ApplyReadoutMitigation:
    name = "readout_base_class"

    @abc.abstractmethod
    def apply_error_mitigation(
        self, results: dict, mapping: Dict, model: QuantumHardwareModel
    ):
        pass

    def process_results(self, results):
        if isinstance(new_results := list(results.values())[0], dict):
            results = new_results
        if list(results.values())[0] > 1:
            shots = sum(list(results.values()))
            results = {key: value / shots for key, value in results.items()}
        return results


class ApplyPostProcReadoutMitigation(ApplyReadoutMitigation):
    name = "post_processing_readout_base_class"

    def apply_error_mitigation(
        self, results: dict, mapping: Dict, model: QuantumHardwareModel
    ):
        raise NotImplementedError()


class ApplyHybridReadoutMitigation(ApplyReadoutMitigation):
    name = "hybrid_readout_base_class"

    def apply_error_mitigation(
        self, results: dict, mapping: Dict, model: QuantumHardwareModel
    ):
        raise NotImplementedError()


class ApplyLinearReadoutMitigation(ApplyPostProcReadoutMitigation):
    name = "linear_readout_mitigation"
    """
    {
        <qubit_number>: {
            "0|0": 1,
            "1|0": 1,
            "0|1": 1,
            "1|1": 1,
        }
    }
    """

    def apply_error_mitigation(
        self, results: dict, mapping: Dict, model: QuantumHardwareModel
    ):
        results = self.process_results(results)
        qubit_count = len(mapping)
        for i in range(qubit_count):
            qubit = mapping[str(i)]
            noise_map = model.error_mitigation.readout_mitigation.linear[str(qubit)]
            results = self.apply_correction_qubit(results, i, noise_map)
        return results

    def apply_correction_qubit(self, results, index, noise_map):
        noise_matrix = np.zeros((2, 2))
        for input_bit in [0, 1]:
            for output_bit in [0, 1]:
                noise_matrix[output_bit, input_bit] = noise_map[f"{output_bit}|{input_bit}"]
        inv_noise_matrix = np.linalg.inv(noise_matrix)
        corrected_results = {}
        for bitstring, probability in results.items():
            corrected_results.setdefault(bitstring, 0.0)
            bit_value = int(bitstring[index])
            other_bit_value = 1 - bit_value
            corrected_results[bitstring] += (
                probability * inv_noise_matrix[bit_value, bit_value]
            )
            other_bitstring = "".join(
                [
                    bitstring[i] if i != index else str(other_bit_value)
                    for i in range(len(bitstring))
                ]
            )
            if other_bitstring in results:
                corrected_results.setdefault(other_bitstring, 0.0)
                corrected_results[other_bitstring] += (
                    probability * inv_noise_matrix[other_bit_value, bit_value]
                )

        # TODO - check validity of ignoring negative probabilities
        corrected_results = {
            key: probability
            for key, probability in corrected_results.items()
            if probability > 0
        }

        return corrected_results

    def __repr__(self):
        return "Linear readout mitigation"


class ApplyMatrixReadoutMitigation(ApplyPostProcReadoutMitigation):
    name = "matrix_readout_mitigation"

    def remap_result(self, results, mapping, output_length):
        output = {}
        for bitstring, result in results.items():
            tmp_bit_string = ["0" for _ in range(output_length)]
            for i, bit in enumerate(bitstring):
                tmp_bit_string[mapping[str(i)]] = bit
            output["".join(tmp_bit_string)] = result
        return output

    def create_array_from_dict(self, algo_dict, n):
        tmp_array = {
            bin(i)[2:].zfill(n): (
                algo_dict[bin(i)[2:].zfill(n)] if algo_dict.get(bin(i)[2:].zfill(n)) else 0
            )
            for i in range(2**n)
        }
        keys = sorted(tmp_array.keys(), key=lambda x: int(x, 2))
        output = []
        for key in keys:
            output.append(tmp_array[key])
        return np.array(output)

    def apply_error_mitigation(self, results: dict, mapping, model: QuantumHardwareModel):
        n = len(mapping)
        results = self.process_results(results)
        algo_data = self.remap_result(results, mapping, n)
        algo_data_array = self.create_array_from_dict(algo_data, n)

        data = np.matmul(
            model.error_mitigation.readout_mitigation.matrix,
            np.transpose(algo_data_array),
        )
        mitigated_data = {bin(i)[2:].zfill(n): data[i] for i in range(2**n)}
        inverted_map = {str(value): int(key) for key, value in mapping.items()}

        return self.remap_result(mitigated_data, inverted_map, n)

    def __repr__(self):
        return "Matrix readout mitigation"


def get_readout_mitigation(mitigation_config: ErrorMitigationConfig):
    mitigators = []
    if ErrorMitigationConfig.LinearMitigation in mitigation_config:
        mitigators.append(ApplyLinearReadoutMitigation())
    if ErrorMitigationConfig.MatrixMitigation in mitigation_config:
        mitigators.append(ApplyMatrixReadoutMitigation())
    return mitigators



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
import json
import re
import sys
from dataclasses import asdict, is_dataclass
from enum import Enum
from importlib import import_module
from json import JSONDecoder, JSONEncoder
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from scc.compiler.hardware_models import QuantumHardwareModel


# Set of common methods so we don't have to add/remove the custom serializer.
def json_dumps(*args, serializable_types=None, **kwargs):
    kwargs.setdefault("cls", CustomJSONEncoder)
    return json.dumps(*args, serializable_types=serializable_types, **kwargs)


def json_loads(
    *args, serializable_types=None, model: "QuantumHardwareModel" = None, **kwargs
):
    kwargs.setdefault("cls", CustomJsonDecoder)
    return json.loads(*args, serializable_types=serializable_types, model=model, **kwargs)


def json_dump(*args, serializable_types=None, **kwargs):
    kwargs.setdefault("cls", CustomJSONEncoder)
    return json.dump(*args, serializable_types=serializable_types, **kwargs)


def json_load(
    *args, serializable_types=None, model: "QuantumHardwareModel" = None, **kwargs
):
    kwargs.setdefault("cls", CustomJsonDecoder)
    return json.load(*args, serializable_types=serializable_types, model=model, **kwargs)


class CustomJsonDecoder(JSONDecoder):
    def __init__(self, *args, serializable_types=None, model=None, **kwargs):
        self.serializable_types = serializable_types
        self.model: "QuantumHardwareModel" = model
        super().__init__(object_hook=self.default, *args, **kwargs)

    def default(self, obj):
        if not isinstance(obj, dict):
            return obj

        # Components are objects directly related to hardware and you never want to serialize them, so we re-link
        # upon deserialization.
        component_id = obj.get("$component_id")
        if component_id is not None:
            if self.model is None:
                raise ValueError(
                    "Attempted to deserialize object that requires re-linking to a "
                    "hardware model and we have no hardware model."
                )

            return self.model.get_device(component_id)

        obj_type = obj.get("$type")
        if obj_type is None:
            return obj

        if obj_type == "<class 'tuple'>":
            return tuple(obj["$data"])

        if self.serializable_types is not None:
            old_paths = ["scc.compiler.config"]
            for old_path in old_paths:
                if old_path in obj_type:
                    obj_type = obj_type.replace(old_path, "qat.purr.compiler.config")
                    break

            typ = self.serializable_types.get(obj_type)
            if typ is None:
                raise ValueError(f"Invalid type attempted to be serialized: {obj_type}.")
        else:
            typ = _get_type(obj_type)

        if issubclass(typ, Enum):
            return typ(obj["$value"])

        if "$data" in obj:
            data = obj["$data"]
            if obj.get("$dataclass", False):
                fields = self.default(data)
                return typ(**fields)
            elif isinstance(data, dict):
                new_obj = object.__new__(typ)
                new_obj.__dict__ = {key: self.default(value) for key, value in data.items()}
                return new_obj
            elif isinstance(data, str):
                return typ(data)
            else:
                return data

        return obj


class CustomJSONEncoder(JSONEncoder):
    """
    It is a customised JSON encoder, which allows the serialization of the more complex
    objects.

    There are four major cases, based on the provided object to be serialized:

    - if the type of the object is supported by the default :class:`JSONEncoder`, than
      the default method is used.
    - if the class of the object is a :class:`dataclass`, then the serialization will
      contain the name of the type, ``dataclass`` flag in order to help at the
      deserialization, and the dictionary of the fields.
    - if the object is none from the above, then the type name is saved, and the
      interior data using ``__dict__``.
    - if an exception is encountered from any cases from above (e.g. ``__dict__`` is not
      available in case of complex numbers), then the type name is saved, and the data
      is the string representation of the object.
    """

    def __init__(self, *args, serializable_types=None, **kwargs):
        self.serializable_types = serializable_types
        super().__init__(*args, **kwargs)

    def default(self, obj):
        typ = type(obj)
        if issubclass(typ, Enum):
            typ_str = f"<enum '{obj.__module__}.{typ.__name__}'>"
        else:
            typ_str = str(typ)

        if self.serializable_types is not None:
            if (
                typ_str not in self.serializable_types
                and type(obj).__module__ != "builtins"
            ):
                raise ValueError(f"Invalid type attempted to be serialized: {(type(obj))}.")

        try:
            from qat.purr.compiler.devices import QuantumComponent
            from qat.purr.compiler.instructions import Acquire

            # TODO: Acquire is a special wrapper component, not an actual component. Have a few too many special-cases
            #   for it now, think about reverting its special status.
            if isinstance(obj, QuantumComponent) and not isinstance(obj, Acquire):
                return {"$component_id": obj.full_id()}
            elif is_dataclass(obj):
                return {
                    "$type": typ_str,
                    "$dataclass": True,
                    "$data": self.default(asdict(obj)),
                }
            elif isinstance(obj, Enum):
                return {"$type": typ_str, "$value": obj.value}
            elif isinstance(obj, complex):
                return {"$type": typ_str, "$data": str(obj)}
            elif isinstance(obj, tuple):
                return {
                    "$type": typ_str,
                    "$data": tuple(self.default(val) for val in obj),
                }

            if hasattr(obj, "__dict__"):
                return {
                    "$type": typ_str,
                    "$data": {
                        key: self.default(value) for key, value in obj.__dict__.items()
                    },
                }
        except (TypeError, AttributeError):
            pass

        return obj


_type_name_matcher = r"<(?:class|enum) '((?:\w+\.)*)(\w+)'>"


def _get_type(s: str):
    match = re.match(_type_name_matcher, s)
    if match:
        namespace = match.group(1)
        if namespace != "":
            expanded_namespace = namespace[:-1].split(".")
            module_name = expanded_namespace.pop(0)
            imported_module = import_module(module_name)
            for item in expanded_namespace:
                module_name = f"{module_name}.{item}"
                try:
                    imported_module = getattr(imported_module, item)
                except AttributeError:
                    import_module(module_name)
                    imported_module = getattr(imported_module, item)
        else:
            imported_module = sys.modules["builtins"]
        try:
            return getattr(imported_module, match.group(2))
        except AttributeError:
            raise AttributeError(
                f"Class {s} not found in built-in modules or namespace '{namespace}'"
            )
    return None



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Oxford Quantum Circuits Ltd
import os.path
import typing
from pathlib import Path
from typing import Union

import regex

import qat.purr.compiler.experimental.frontends as experimental_frontends
import qat.purr.compiler.frontends as core_frontends
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.config import CompilerConfig
from qat.purr.compiler.frontends import LanguageFrontend
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.metrics import CompilationMetrics
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()

# Credible input are str (contents or file path) or pre-build instruction builder
QATInput = Union[str, bytes, InstructionBuilder]


def _return_or_build(qat_input: QATInput, build_func: typing.Callable, **kwargs):
    if isinstance(qat_input, (str, bytes)):
        return build_func(qat_input, **kwargs)

    if isinstance(qat_input, InstructionBuilder):
        return qat_input, CompilationMetrics()

    raise TypeError(f"No compiler support for inputs of type {str(type(qat_input))}")


def execute(
    qat_input: QATInput,
    hardware: QuantumHardwareModel = None,
    compiler_config: CompilerConfig = None,
):
    """Execute file path or code blob."""
    results, _ = execute_with_metrics(qat_input, hardware, compiler_config)
    return results


contents_match_pattern = regex.compile(
    '(OPENQASM [0-9]*(.0)?;|defcalgrammar "[a-zA-Z ]+";)|(@__quantum__qis)'
)

path_regex = regex.compile(r"^.+\.(qasm|ll|bc)$")


def fetch_frontend(
    path_or_str: Union[str, bytes],
    use_experimental: bool = False,
) -> LanguageFrontend:
    frontend_mod = core_frontends
    if use_experimental:
        frontend_mod = experimental_frontends
    if isinstance(path_or_str, bytes) and path_or_str.startswith(b"BC"):
        return frontend_mod.QIRFrontend()

    if path_regex.match(path_or_str) is not None:
        if not os.path.exists(path_or_str):
            raise ValueError(f"Path {path_or_str} does not exist.")

        path = Path(path_or_str)
        if path.suffix == ".qasm":
            return frontend_mod.QASMFrontend()
        elif path.suffix in (".bc", ".ll"):
            return frontend_mod.QIRFrontend()

        raise ValueError(f"File with extension {path.suffix} unrecognized.")

    results = regex.search(contents_match_pattern, path_or_str)
    if results is not None:
        if results.captures(1):
            return frontend_mod.QASMFrontend()
        if results.captures(3):
            return frontend_mod.QIRFrontend()
    else:
        raise ValueError("Cannot establish a LanguageFrontend based on contents")


def execute_with_metrics(
    path_or_str: str,
    hardware: QuantumHardwareModel = None,
    compiler_config: CompilerConfig = None,
):
    """Execute file path or code blob."""

    frontend: LanguageFrontend = fetch_frontend(path_or_str)
    return _execute_with_metrics(frontend, path_or_str, hardware, compiler_config)


def execute_qir(qat_input: QATInput, hardware=None, compiler_config: CompilerConfig = None):
    results, _ = execute_qir_with_metrics(qat_input, hardware, compiler_config)
    return results


def execute_qir_with_metrics(
    qat_input: QATInput, hardware=None, compiler_config: CompilerConfig = None
):
    frontend = core_frontends.QIRFrontend()
    return _execute_with_metrics(frontend, qat_input, hardware, compiler_config)


def execute_qasm(
    qat_input: QATInput, hardware=None, compiler_config: CompilerConfig = None
):
    results, _ = execute_qasm_with_metrics(qat_input, hardware, compiler_config)
    return results


def execute_qasm_with_metrics(
    qat_input: QATInput, hardware=None, compiler_config: CompilerConfig = None
):
    frontend = core_frontends.QASMFrontend()
    return _execute_with_metrics(frontend, qat_input, hardware, compiler_config)


def _execute_with_metrics(
    frontend: LanguageFrontend,
    qat_input: QATInput,
    hardware=None,
    compiler_config: CompilerConfig = None,
):
    metrics = CompilationMetrics()
    if compiler_config is not None:
        metrics.enable(compiler_config.metrics, overwrite=True)

    instructions, build_metrics = _return_or_build(
        qat_input, frontend.parse, hardware=hardware, compiler_config=compiler_config
    )
    metrics.merge(build_metrics)

    results, execution_metrics = frontend.execute(instructions, hardware, compiler_config)
    metrics.merge(execution_metrics)

    return results, metrics.as_dict()
